/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSPeerSession_cxx_Version = 
"$Id: LSPeerSession.cxx,v 1.17 2000/11/13 21:48:44 kevin Exp $";

#include <iostream.h>
#include <strstream.h>
#include "LSPeerSession.hxx"
#include "TcpClientSocket.hxx"
#include "TRIPTransceiver.hxx"
#include "TRIPOpenMsg.hxx"
#include "TRIPUpdateMsg.hxx"
#include "TRIPStateUnknown.hxx"
#include "TRIPNotificationMsg.hxx"
#include "TRIPKeepAliveMsg.hxx"
#include "LSLocator.hxx"
#include "LockHelper.hxx"
#include "cpLog.h"

LSPeerSession::LSPeerSession(u_int16_t itad,
                 const TRIPId& tripId,const NetworkAddress& lsAddress)
{
    _parentItad = itad;
    _tripId = tripId;
    _peerAddress = new NetworkAddress(lsAddress);
    _destIp = _peerAddress->getIpName();

    _state = new TRIPStateUnknown();
    _state->init(*this);
    _thread = new VThread();
    _thread->spawn(LSPeerSession::run, this);
}

LSPeerSession::LSPeerSession(u_int16_t itad, u_int16_t pItad,
                      const TRIPId& tripId, const TRIPId& pTripId,                                             const Connection& conn)
{
    _parentItad = itad;
    _peerParentItad = pItad;
    _tripId = tripId;
    _peerTripId = pTripId;
    _connection = conn;
    _destIp = _connection.getIp();

    _state = new TRIPStateUnknown();
    _state->init(*this);

    _thread = new VThread();
    _thread->spawn(LSPeerSession::run, this);
}

LSPeerSession::~LSPeerSession()
{
    if(_thread != 0) _thread->exit();
}

void
LSPeerSession::start()
{
    try
    {
        _state->start(*this);
        connect();
        _state->sendOpen(*this);
    }
    catch(VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Network error, reason:%s",
                          e.getDescription().c_str());
        _state->lostConnection(*this);
    }
}


void
LSPeerSession::connect()
{
     LockHelper lck(_mutex);
    if(_connection.isLive())
    {
        cpLog(LOG_INFO, "Already connected to (%s)", 
                                 _connection.getDescription().c_str());        
        return;
    }
    TcpClientSocket cSock(*_peerAddress, false);
   
    cSock.connect();
    _connection = cSock.getConn();

    assert(_connection.isLive());
    cpLog(LOG_INFO, "Connected to (%s)", _connection.getDescription().c_str());
}

void 
LSPeerSession::dump(ostream& os) const
{
    os << "(" << _parentItad << "," << _tripId << ")<->(" ;
    os << _peerParentItad << "," << _peerTripId << ")" << endl; 
}

bool 
LSPeerSession::peerExists(u_int16_t itad, const TRIPId& tripId)
{
    if((_peerParentItad == itad) && (_peerTripId == tripId))
    {
        return true;
    }
    return false;
}

bool 
LSPeerSession::peerExists(const string& destIp)
{
    if(_destIp == destIp)
    {
        return true;
    }
    return false;
}

void*
LSPeerSession::run(void* args)
{
    LSPeerSession* self = (LSPeerSession*)args;
    self->processMsg();
    return (NULL);
}

void
LSPeerSession::processMsg()
{
    for (; ; )
    {
        if(_state->isEstablished())
        {
            //cpLog(LOG_DEBUG, "Processing message in state:%s", _state->description().c_str());
        }
        LockHelper lck(_mutex);
        if(_connection.isLive() && _connection.isReadReady(1))
        {
            TRIPTransceiver trans;
            Sptr<TRIPMsg> msg = 0;
            try
            {
                msg = trans.receiveMsg(_connection.getConnId());
                if(msg->isAOpenMsg())
                {
                    processOpenMsg(msg);
                }
                else if(msg->isAKeepAliveMsg())
                {
                   
                    //cpLog(LOG_DEBUG, "Processing message in state:%s", _state->description().c_str());
                    _state->receiveKeepAlive(*this);
                }
                else if(msg->isANotificationMsg())
                {
                    _state->receiveNotification(*this);
                }
                else if(msg->isAUpdateMsg())
                {
                    handleUpdateMessage(msg);
                }
                //Enabling this would demonstrate UPDATE message 
                //being exchanged between two Peers
                //advertiseRoutes();
            }
            catch(VNetworkException& e)
            {
                cpLog(LOG_ALERT, "Network error, reason:%s",
                                     e.getDescription().c_str());
                _state->lostConnection(*this);
            }
            catch(TRIPInvalidStateException& e)
            {
                cpLog(LOG_ALERT, "Invalid state transition , reason:%s",
                                     e.getDescription().c_str());
            }
            catch(TRIPBadDataException& e)
            {
                cpLog(LOG_ALERT, "Bad TRIP data, reason:%s",
                                     e.getDescription().c_str());
                _state->lostConnection(*this);
            }
        }
        else if(!_connection.isLive())
        {
            lck.unlock();
            usleep(100000);
        }
    }
}

void
LSPeerSession::processOpenMsg(Sptr<TRIPMsg> msg) throw (VNetworkException&, TRIPBadDataException&)
{
    cpLog(LOG_DEBUG, "processOpenMsg");
    if(msg != 0)
    {
        Sptr<TRIPOpenMsg> openMsg;
        openMsg.dynamicCast(msg);
        const TRIPOpenMsg::ParameterList& pList = openMsg->getParameters();
        assert(pList.size());
        TRIPOpenMsg::ParameterList::const_iterator itr;
        for (itr = pList.begin(); itr != pList.end(); itr++)
        {
            const Sptr<TRIPParameter>& param = (*itr);
            if(param->getType() == TRIP_PARAM_CAP)
            {
                Sptr<TRIPCapabilityInfo> cInfo;
                cInfo.dynamicCast(param); 
                assert(cInfo.getPtr());
                const TRIPCapabilityInfo::CapabilityList& cList = cInfo->getData();
                TRIPCapabilityInfo::CapabilityList::const_iterator cItr;
                for (cItr = cList.begin(); cItr != cList.end(); cItr++)
                {
                    const Sptr<TRIPCapability>& cap = (*cItr);
                    cap->validate();
                    if(cap->getCode() == TRIP_CAP_SND_REC)
                    {
                        Sptr<TRIPTransCapability> transCap ; 
                        transCap.dynamicCast(cap);
                        assert(transCap.getPtr());
                        transCap->validate();

                        if((transCap->isSendOnly() && (getMode() == TRIP_TRANS_SND)) ||
                           (transCap->isRecvOnly() && (getMode() == TRIP_TRANS_RCV)))
                        {
                            //Send the notification and close the connection
                            TRIPData data("Unsupported capability");
                            TRIPNotificationMsg msg(TRIP_ERR_OPEN, 6, data);
                            msg.writeData(_connection.getConnId());
                            _state->fail(*this);
                            return;
                        }
                        try
                        {
                            _state->receiveOpen(*this);
                            sendKeepAlive();
                            _holdTimeAgent->setDuration(openMsg->getHoldTime()*1000);
                            _keepAliveAgent->setDuration((openMsg->getHoldTime()*1000)/3);
                        }
                        catch(TRIPInvalidStateException& e)
                        {
                            cpLog(LOG_DEBUG, "Invalid state transition, (%s)",
                                             e.getDescription().c_str());
                        }
                        catch(VNetworkException& e)
                        {
                            cpLog(LOG_DEBUG, "Network error, (%s)",
                                             e.getDescription().c_str());
                            _state->fail(*this);
                        }
                    }
                }
            }
        }
    }
}

ostream&
operator<<(ostream& os, const LSPeerSession& lsSession)
{
    lsSession.dump(os);
    return os;
}

void
LSPeerSession::sendNotification(u_int8_t eCode, u_int8_t eSubCode, const string& msg)
{
    try
    {
        TRIPData data(msg);
        TRIPNotificationMsg msg(eCode, eSubCode, data);
        if(_connection.isLive())
        {
            msg.writeData(_connection.getConnId());
        }
    }
    catch(VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Failed to send notification msg");
    }
}

void 
LSPeerSession::sendOpen() throw (VNetworkException&)
{
    // TODO Need to cancel ConnectRetryTimer, complete init??, 
    // send OPEN msg, set Hold Timer to 4mins/whatever..
    assert(0);
}

void
LSPeerSession::sendKeepAlive()
{
    try
    {
        TRIPKeepAliveMsg msg;
        msg.writeData(_connection.getConnId());
    }
    catch(VException& e)
    {
        cpLog(LOG_INFO, "Got exception (%s), in state (%s)",
                         e.getDescription().c_str(), _state->description().c_str());
        _state->fail(*this);
    }
}

void
LSPeerSession::holdTimerExpired() 
{
    _state->holdTimerExpired(*this);
}

void
LSPeerSession::restartTimerExpired() 
{
    try
    {
        _state->start(*this);
        connect();
        _state->sendOpen(*this);
    }
    catch(VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Network error, reason:%s",
                          e.getDescription().c_str());
        _state->lostConnection(*this);
    }
}

void
LSPeerSession::connectRetryTimerExpired() 
{
    _state->connectRetryTimerExpired(*this);
}

void
LSPeerSession::setConnectRetryTimer()
{
    cpLog(LOG_DEBUG, "LSPeerSession::setConnectRetryTimer");
    _connectRetryAgent->startTimer();
}

void
LSPeerSession::setHoldTimer()
{
    _holdTimeAgent->startTimer();
}

void
LSPeerSession::setRestartTimer()
{
    _restartAgent->startTimer();
}


void
LSPeerSession::cancelConnectRetryTimer()
{
    cpLog(LOG_DEBUG, "LSPeerSession::cancelConnectRetryTimer");
    _connectRetryAgent->cancelTimer();
}

void
LSPeerSession::cancelHoldTimer()
{
    _holdTimeAgent->cancelTimer();
}


void
LSPeerSession::initKeepAlive()
{
    _keepAliveAgent->startTimer();
}

void
LSPeerSession::stopKeepAlive()
{
    _keepAliveAgent->cancelTimer();
}



void
LSPeerSession::makeConnection()
{
    connect();
}

void
LSPeerSession::closeConnections()
{
    if(_connection.isLive())
    {
        _connection.close();
    }
}

void
LSPeerSession::initTRIPResources()
{
    // TODO Initialize required resources
    // Need to create the Timer Agents
    // Need to set all of the initial timer durations if required.
    _keepAliveAgent = new KeepAliveAgent(*this);
    _holdTimeAgent = new HoldTimeAgent(*this);
    _connectRetryAgent = new ConnectRetryAgent(*this);
    _restartAgent = new RestartAgent(*this);

    // Initial timer durations - may be changed later after peer negotiation
    _keepAliveAgent->setDuration(80000); //1/3 of holdTimer
    _holdTimeAgent->setDuration(240000); // recommended initial 4 mins
    _connectRetryAgent->setDuration(30000); // locally set time - 1/2 min
    _restartAgent->setDuration(60000); // recommended 1 min initially
}

void
LSPeerSession::releaseTRIPResources()
{
    _keepAliveAgent->cancelTimer();
    _holdTimeAgent->cancelTimer();
    _connectRetryAgent->cancelTimer();
    _restartAgent->cancelTimer();
    _connectRetryAgent->setDuration(30000); // reset to initial value
}


void 
LSPeerSession::handleIncomingConnection(Connection& conn)
{
    LockHelper lck(_mutex);
    if(_state->acceptAllowed())
    {
        if(!LSLocator::instance().isTrustedLs(conn.getIp()))
        {
            cpLog(LOG_DEBUG, "LS is not trusted, ignoring the connection");
            conn.close();
            return;
        }
        cpLog(LOG_DEBUG, "LSPeerSession::handleIncomingConnection, acceptAllowed");
        if(_connection.isLive())
        {
            cpLog(LOG_DEBUG, "Closing connection (%s)", _connection.getDescription().c_str());
            _connection.close();
        }
        _connection = conn;
        cpLog(LOG_DEBUG, "Assigned connection (%s)", _connection.getDescription().c_str());
        _state->sendOpen(*this);
    }
    else
    {
        cpLog(LOG_DEBUG, "LSPeerSession::handleIncomingConnection, accept not Allowed");
        cpLog(LOG_DEBUG, "Closing connection (%s)", conn.getDescription().c_str());
        conn.close();
    }
}

void
LSPeerSession::handleUpdateMessage(Sptr<TRIPMsg> msg)
{
    Sptr<TRIPUpdateMsg> uMsg;
    uMsg.dynamicCast(msg);
    if(uMsg != 0)
    {
        //TODO
    }
}

void
LSPeerSession::advertiseRoutes()
{
    //TESTING ONLY
    //Create the UPdate Message
    TRIPUpdateMsg uMsg;
    Sptr<TRIPWithdrawnRoute> route1 = new TRIPWithdrawnRoute(TRIP_FLG_TRANSITIVE);
    Sptr<TRIPReachableRoute> route2 = new TRIPReachableRoute(TRIP_FLG_TRANSITIVE);
    TRIPRoutes tRoute(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP,TRIPData("1408"));
    TRIPRoutes tRoute2(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP,TRIPData("1508"));
    *route1 << tRoute;
    *route2 << tRoute2;

    uMsg.addWithdrawnRoute(route1);
    uMsg.addReachableRoute(route2);
    
    TRIPItadPathSegment seg1(TRIP_SEG_AP_SET, 1224);
    Sptr<TRIPAdvertisePath> aPath = new TRIPAdvertisePath(TRIP_FLG_TRANSITIVE);
    *aPath << seg1;
    
    TRIPItadPathSegment seg2(TRIP_SEG_AP_SET, 1224);
    Sptr<TRIPRoutedPath> aPath2 = new TRIPRoutedPath(TRIP_FLG_TRANSITIVE);
    *aPath2 << seg2;

    uMsg.addAdvertisePath(aPath); 
    uMsg.addRoutedPath(aPath2); 

    Sptr<TRIPNextHopServer> nServer = new TRIPNextHopServer(TRIP_FLG_TRANSITIVE, 1223,
                                                      TRIPData("129.234.34.56"));
    uMsg.addAttribute(nServer);
    cpLog(LOG_DEBUG,  "Sending Update Msg");
    sendUpdate(uMsg);
}

void 
LSPeerSession::sendUpdate(const TRIPUpdateMsg& uMsg) throw (VNetworkException&)
{
    if(_connection.isLive())
    {
        uMsg.writeData(_connection.getConnId());
    }
}

