#ifndef LSPeerSession_hxx
#define LSPeerSession_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSPeerSession_hxx_Version = 
"$Id: LSPeerSession.hxx,v 1.11 2000/11/08 23:51:41 surendra Exp $";

#include "global.h"
#include <vector>
#include <string>

#include "vtypes.h"
#include "Sptr.hxx"
#include "NetworkAddress.h"
#include "TRIPId.hxx"
#include "TRIPState.hxx"
#include "Connection.hxx"
#include "VThread.h"
#include "TRIPTransCapability.hxx"
#include "VNetworkException.hxx"
#include "TRIPTimerAgent.hxx"
#include "KeepAliveAgent.hxx"
#include "HoldTimeAgent.hxx"
#include "RestartAgent.hxx"
#include "ConnectRetryAgent.hxx"
#include "VMutex.h"

class TRIPMsg;
class TRIPUpdateMsg;

/** 
    Base class LSPeerSession implements  the peer-to-peer connection between two LSs
    and drive the FSM for that connection.It uses TRIPState object to keep track of
    the connection states.Based on the capabilities any of the derivatives
    LSSendOnlySession, LSReceiveOnlySession and LSSendReceiveSession objects are
    created.
*/


class LSPeerSession {

    public:
        ///
        LSPeerSession(u_int16_t itad, 
                      const TRIPId& tripId, const NetworkAddress& lsAddress);

        ///
        LSPeerSession(u_int16_t itad, u_int16_t pItad,
                      const TRIPId& tripId, const TRIPId& pTripId, 
                      const Connection& conn);

        ///
        virtual string className() { return "LSPeerSession"; }

        ///Destructor
        virtual ~LSPeerSession();

        ///
        friend ostream& operator<<(ostream& os, const LSPeerSession& lssession);

        ///
        virtual void dump(ostream& os) const;

        ///
        virtual TRIPTransMode getMode() const = 0;

        ///Returns true if it matches the given itad and trip ID
        bool peerExists(u_int16_t itad, const TRIPId& tripId);

        ///Returns true if it matches the given destIp
        bool peerExists(const string& destIp);

        ///Sends a update message to peer
        virtual void sendUpdate(const TRIPUpdateMsg& uMsg) throw (VNetworkException&);

        ///
        void handleIncomingConnection(Connection& conn);

  	///
	void sendNotification(u_int8_t eCode, u_int8_t eSubCode, const string& msg);

	///
	void sendKeepAlive();

 	///
	void holdTimerExpired() ;

	///
	void restartTimerExpired() ;

        ///
        void connectRetryTimerExpired() ;

	///
	void setConnectRetryTimer();

	///
	void cancelConnectRetryTimer();

	///
	void setHoldTimer();

        ///
        void cancelHoldTimer();

	///
	void setRestartTimer();

	///
	void initKeepAlive();

      	///
	void stopKeepAlive();

	///
	void makeConnection();

 	///
	void closeConnections();

	///
	virtual void initTRIPResources();

	///
	virtual void releaseTRIPResources();

        ///
        void start();

    protected:

        void processOpenMsg(Sptr<TRIPMsg> msg) throw (VNetworkException&, TRIPBadDataException&);


 	///
	void setState(TRIPState* newState)
	{
	    cpLog(LOG_DEBUG, "LSPeerSession::setState - oldstate %s",
				_state->description().c_str());
	    _state = newState;
	    cpLog(LOG_DEBUG, "LSPeerSession::setState - newstate %s",
				_state->description().c_str());
	};

        ///
        virtual void sendOpen() throw (VNetworkException&);

        ///
        virtual void handleUpdateMessage(Sptr<TRIPMsg> msg);

        ///Get the routes to be advertise and asend update message to Peer
        virtual void advertiseRoutes();


        ///
        void connect();
        ///
        static void* run(void* args);

        ///
        void processMsg();

        ///
        Sptr <NetworkAddress> _peerAddress;

        ///
        Connection     _connection;
 
	///
	string 	_destIp;

        ///
        u_int16_t   _parentItad;
        ///
        u_int16_t   _peerParentItad;

        ///
        TRIPId      _tripId;
        ///
        TRIPId      _peerTripId;

        Sptr<VThread>  _thread;

        ///
        mutable Sptr<TRIPState> _state;

	///
	Sptr <KeepAliveAgent> _keepAliveAgent;

	///
	Sptr <HoldTimeAgent> _holdTimeAgent;

	///
	Sptr <RestartAgent> _restartAgent;

   	///
	Sptr <ConnectRetryAgent> _connectRetryAgent;

        ///
        VMutex       _mutex;

        ///
        friend class TRIPState;
        friend class TRIPStateConnect;
        friend class TRIPStateActive;
};


#endif
