/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSSessionManager_cxx_Version = 
"$Id: LSSessionManager.cxx,v 1.9 2000/11/13 23:41:58 surendra Exp $";


#include <unistd.h>
#include <iostream>
#include <stdlib.h> 
#include <strstream.h>
#include "LSLocator.hxx"
#include "LSSessionManager.hxx"
#include "LSSendOnlySession.hxx"
#include "LSReceiveOnlySession.hxx"
#include "LSSendReceiveSession.hxx"
#include "TcpServerSocket.hxx"
#include "TRIPOpenMsg.hxx"
#include "TRIPTransceiver.hxx"
#include "cpLog.h"

LSSessionManager* LSSessionManager::_instance = 0;

LSSessionManager&
LSSessionManager::instance(u_int16_t itad, TRIPTransMode tMode)
{
    if (!_instance)
    {
        _instance = new LSSessionManager(itad, tMode);
    }
    return (*_instance);
}

void 
LSSessionManager::destroy(void)
{
    delete LSSessionManager::_instance;
    LSSessionManager::_instance = 0;
}

LSSessionManager::LSSessionManager(u_int16_t itad, TRIPTransMode tMode)
{
    _itad = itad;
    _transMode = tMode;

    char buf[256];
    strcpy(buf, "localhost");
    gethostname(buf, 256);
    NetworkAddress na(buf);
    _tripId = TRIPId(na.getIp4Address ()); 

    int port = 6069;
#ifdef __linux__
    if (atexit(LSSessionManager::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };
#endif

    _serverSocket = new TcpServerSocket(port);

    init();
}

LSSessionManager::~LSSessionManager()
{
    if(_thread != 0)
    {
        _thread->exit();
    }
}

void
LSSessionManager::run()
{
    ///If threadRun is already running, call to run
    //should be discarded
    if(_thread.getPtr() == 0)
    {
        processMsg();
    }
    else
    {
        cpLog(LOG_ALERT, "Already processing messages in separate thread");
    }
}

void
LSSessionManager::threadRun()
{
    _thread = new VThread();
    _thread->spawn(LSSessionManager::threadRunImpl, this);
}

void*
LSSessionManager::threadRunImpl(void* args)
{
    LSSessionManager* self = (LSSessionManager*)args;
    self->processMsg();
    return (NULL);
}

void
LSSessionManager::processMsg()
{
    for (; ; )
    {
        cpLog(LOG_DEBUG, "Accepting connection");
        Connection conn;
        _serverSocket->accept(conn);
        try
        {
             if(!LSLocator::instance().isTrustedLs(conn.getIp()))
             {
                 cpLog(LOG_DEBUG, "LS is not trusted, ignoring the connection");
                 conn.close();
                 continue;
             }
             processOpenRequest(conn);
        }
        catch(TRIPBadDataException& e)
        {
            cpLog(LOG_ALERT, "Bad open message, reason:%s",
                             e.getDescription().c_str());
        }
        catch(VException& e)
        {
            cpLog(LOG_ALERT, "Caught exception, reason:%s",
                             e.getDescription().c_str());
        }
    }
}

void
LSSessionManager::processOpenRequest(Connection& conn) throw (VNetworkException&, TRIPBadDataException&)
{
    string destIp = conn.getIp();
    cpLog(LOG_DEBUG, "processOpenRequest from IP:%s", destIp.c_str());
    Sptr<LSPeerSession> pSession = locateExistingPeerSession(destIp);
    if(pSession != 0)
    { 
        cpLog(LOG_DEBUG, "Found existing PeerSession for (%s)", destIp.c_str());
        pSession->handleIncomingConnection(conn);
    }
    else
    {
        cpLog(LOG_DEBUG, "Creating new PeerSession for (%s)", destIp.c_str());
        NetworkAddress addr(conn.getIp(), 6069);
        if(_transMode == TRIP_TRANS_SND)
        {
            pSession  = new LSSendOnlySession(_itad, _tripId, addr);
        }
        else if (_transMode == TRIP_TRANS_RCV)
        {
            pSession  = new LSReceiveOnlySession(_itad, _tripId, addr);
        }
        else
        {
            pSession  = new LSSendReceiveSession(_itad, _tripId, addr);
        }
        _sessionList.push_back(pSession);
        pSession->handleIncomingConnection(conn);
        pSession->start();
    }
}

void 
LSSessionManager::init()
{
    ///Find out all LSs out there and intiate a session
    //TODO
    const LSLocator::LSList& lsList = LSLocator::instance().getLSList();
    for(LSLocator::LSList::const_iterator itr = lsList.begin();
                itr != lsList.end(); itr++)
    {
        NetworkAddress addr((*itr));
        Sptr<LSPeerSession> sSession;
        if(_transMode == TRIP_TRANS_SND)
        {
            sSession  = new LSSendOnlySession(_itad, _tripId, addr);
        }
        else if(_transMode == TRIP_TRANS_RCV)
        {
            sSession  = new LSReceiveOnlySession(_itad, _tripId, addr);
        }
        else
        {
            sSession = new LSSendReceiveSession(_itad, _tripId, addr);
        }
        _sessionList.push_back(sSession);
        sSession->start();
    }
}

ostream&
operator<<(ostream& os, const LSSessionManager& sessionManager)
{
    LSSessionManager::LSPeerList::const_iterator itr;
    os << "LS Session list:" << endl;
    for(itr = sessionManager._sessionList.begin(); 
              itr != sessionManager._sessionList.end(); itr++)
    {
        os << *(*itr) << endl;
    }
    return os;
}

Sptr<LSPeerSession> 
LSSessionManager::locateExistingPeerSession(const string& destIp)
{
    Sptr<LSPeerSession> retVal;
    LSPeerList::const_iterator itr;
    for(itr = _sessionList.begin(); 
              itr != _sessionList.end(); itr++)
    {
        if((*itr)->peerExists(destIp))
        {
            retVal = (*itr);
            break;
        }
    }
    return retVal;
}

bool 
LSSessionManager::locateExistingPeerSession(u_int16_t itad, const TRIPId& tripId)
{
    bool retVal = false;
    LSPeerList::const_iterator itr;
    for(itr = _sessionList.begin(); 
              itr != _sessionList.end(); itr++)
    {
        if((*itr)->peerExists(itad, tripId))
        {
            retVal = true;
            break;
        }
    }
    return retVal;
}
