/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const TRIPNotificationMsg_cxx_Version = 
"$Id: TRIPNotificationMsg.cxx,v 1.2 2000/11/09 01:26:57 surendra Exp $";

#include "TRIPNotificationMsg.hxx"
#include "TRIPUtil.hxx"
#include "cpLog.h"

TRIPNotificationMsg::TRIPNotificationMsg(u_int8_t eCode, 
                         u_int8_t eSubCode,const TRIPData& data)
    :TRIPMsg(), 
     _errCode(eCode), 
     _errSubCode(eSubCode),
     _data(data) 
{
    //Length of message 
    //3 bytes for header + 2 bytes for(error code and subcode) + dataLength
    _hdr = new TRIPHeader(TRIP_OP_NOTIF, _data.length()+5);
    checkSanity();
}

void 
TRIPNotificationMsg::dump(ostream& os)const
{
   os.width(25);
   os << "**NotificationMsg**" << endl;
   os << *_hdr << endl;
   os.width(25);
   os << "Error code:" << static_cast<int>(_errCode) << endl;
   os.width(25);
   os << "Error subcode:" << static_cast<int>(_errSubCode) << endl;
   os.width(25);
   os << "Data:" << _data << endl;
}

void 
TRIPNotificationMsg::parse(const TRIPHeader& hdr,
                          const char* data) throw (TRIPBadDataException&)
{
    if(!(hdr.isANotificationHdr()) && (hdr.getMsgLength() < 5))
    {
        cpLog(LOG_ALERT, "Bad Notification Msg");
        throw TRIPBadDataException("Bad Notification Msg", __FILE__, __LINE__, 0);
    }

    _hdr = new TRIPHeader(hdr);
 
    int idx = 0;
    memcpy(&_errCode, data , 1); 
    idx += 1;
    memcpy(&_errSubCode, data+idx , 1);
    idx += 1;

    int dLen = _hdr->getMsgLength() - 5;

    _data = TRIPData(data+idx, dLen);

    setMsgLength(_data.length() + 5);
}

void 
TRIPNotificationMsg::parse(const char* data) throw (TRIPBadDataException&) 
{
    TRIPHeader hdr(data);
    parse(hdr, data+3);
}


void 
TRIPNotificationMsg::writeData(int id) const throw(VNetworkException&)
{
   TRIPMsg::writeData(id);

   struct shipIt {
       u_int8_t    eCode;
       u_int8_t    eSubCode;
   } sIt;
   sIt.eCode = _errCode;
   sIt.eSubCode = _errSubCode;

   TRIPUtil::writeData(id, &sIt, sizeof(struct shipIt));
   if(_data.length())
   {
       TRIPUtil::writeData(id, _data.getData(), _data.length());
   }
}

void 
TRIPNotificationMsg::checkSanity() const throw (TRIPBadDataException&)
{
    if(!(_hdr->isANotificationHdr()) && (_hdr->getMsgLength() < 5))
    {
        cpLog(LOG_ALERT, "Bad Notification Msg");
        throw TRIPBadDataException("Bad Notification Msg", __FILE__, __LINE__, 0);
    }

    //Validate the error code and errorSubode
    bool failed = false;
    switch(static_cast<TRIPErrorCode>(_errCode))
    {
        case TRIP_ERR_MSG_HEADER:
        {
            if(_errSubCode < 1 && _errSubCode > 2)
            {
                failed = true;
            }
        }
        break;
        case TRIP_ERR_OPEN:
        case TRIP_ERR_UPDATE:
        {
            if(_errSubCode < 1 && _errSubCode > 6)
            {
                failed = true;
            }
        }
        break;
        case TRIP_ERR_HOLD_TIMER:
        case TRIP_ERR_FSM:
        case TRIP_ERR_CEASE:
        {
            if(_errSubCode  > 0)
            {
                failed = true;
            }
        }
        break;
        default:
        {
            failed = true;
        }
        break;
    }

    if(failed)
    {
        cpLog(LOG_ALERT, "Bad Notification Msg");
        throw TRIPBadDataException("Bad Notification Msg", __FILE__, __LINE__, 0);
    }
     
}

