#ifndef CWBUFFER_HXX_
#define CWBUFFER_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const CWBuffer_hxx_Version =
    "$Id: CWBuffer.hxx,v 1.7 2000/09/20 08:23:35 cvs Exp $";

#include "CountSemaphore.hxx"
#include <string.h>
#include <ctype.h>
#include "Sptr.hxx"
#include "fastmemcpy.hxx"

#define INLINE_ inline

INLINE_ int
mystrlen(const char* s)
{
    int i = 0;
    while (*s++ != '\0')
    {
        i++;
    }
    return i;
}

INLINE_ void*
mmemcpy(char* dest, const char* s, int len)
{
    char *d = dest;
    while (len--)
    {
        *d++ = *s++;
    }
    return dest;
}

static char* emptyStr = "";


struct CWCharArray
{
    CWCharArray(int length);
    CWCharArray(const char* p, int len, int allocsize);
    ~CWCharArray();

    CountSemaphore mutex;
    char* ptr;
};

INLINE_
CWCharArray::CWCharArray(int length)
        : mutex()
{
    ptr = new char[length];
}


INLINE_
CWCharArray::CWCharArray(const char* p, int len, int allocsize)
        : mutex()
{
    assert(allocsize >= len);

    ptr = new char[allocsize];
    mmemcpy(ptr, p, len);
}


INLINE_
CWCharArray::~CWCharArray()
{
    assert(mutex.compare(0));
    delete[] ptr;
}


INLINE_ void decrementCA(CWCharArray* ptr)
{
    if (ptr)
    {
        if (ptr->mutex.decrement())
        {
            delete (ptr);
        }
    }
}



struct CWBuffer
{
    CWBuffer(int length = 0);
    CWBuffer(const char* src, int length, int hint = 0);
    CWBuffer(const CWBuffer& src);
    ~CWBuffer();

    const CWBuffer& operator=(const CWBuffer& src);

    // these functions don't make modifications to the underlying
    // buffer, so they ought to be cheap

    void clear();
    int compare(const char* str, int rightSize) const;
    int compareNoCase(const char* str, int rightSize) const;
    void split(int p, CWBuffer* ptr);
    void truncate(int front, int back);
    void substr(int first, int last);

    // these functions (potentially) modify the underlying buffer, so
    // they are (potentially) expensive

    void makeWritable(int length);
    void makeCString();


    // the data goes here

    CWCharArray* bufPtr;
    char* bitsPtr;
    int size;
    int allocSpace;
};


INLINE_
CWBuffer::CWBuffer(int length)
        :
        bufPtr(0),
        bitsPtr(emptyStr),
        size(0),
        allocSpace(length)
{
    if (allocSpace)
    {
        bufPtr = new CWCharArray(allocSpace);
        bufPtr->mutex.increment();
        bitsPtr = bufPtr->ptr;
    }
}


INLINE_
CWBuffer::CWBuffer(const char* src, int length, int hint)
        :
        bufPtr(0),
        bitsPtr(emptyStr),
        size(length),
        allocSpace(length)
{
    if (hint > length)
    {
        allocSpace = hint;
    }

    if (allocSpace)
    {
        bufPtr = new CWCharArray(src, length, allocSpace);
        bufPtr->mutex.increment();
        bitsPtr = bufPtr->ptr;
    }
}


INLINE_
CWBuffer::CWBuffer(const CWBuffer& src)
        :
        bufPtr(src.bufPtr),
        bitsPtr(src.bitsPtr),
        size(src.size),
        allocSpace(src.allocSpace)
{
    if (bufPtr)
    {
        bufPtr->mutex.increment();
    }
}


INLINE_ const CWBuffer&
CWBuffer::operator=(const CWBuffer& src)
{
    bufPtr = src.bufPtr;
    bitsPtr = src.bitsPtr;
    size = src.size;
    allocSpace = src.allocSpace;

    if (bufPtr)
    {
        bufPtr->mutex.increment();
    }
    return *this;
}


INLINE_ void
CWBuffer::clear()
{
    if (bufPtr)
    {
        decrementCA(bufPtr);
    }
    bitsPtr = emptyStr;
    size = 0;
    allocSpace = 0;
    bufPtr = 0;
}



INLINE_
CWBuffer::~CWBuffer()
{
    decrementCA(bufPtr);
}


INLINE_ int
CWBuffer::compare(const char* str, int rightSize) const
{
    char* ptr = bitsPtr;
    int leftSize = size;

    while (leftSize && rightSize)
    {
        if (*ptr < *str)
        {
            return -1;
        }
        if (*ptr > *str)
        {
            return 1;
        }
        leftSize--;
        rightSize--;
        ptr++;
        str++;
    }

    if (leftSize < rightSize)
    {
        return -1;
    }
    else if (leftSize > rightSize)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}


INLINE_ int
CWBuffer::compareNoCase(const char* str, int rightSize) const
{
    char* ptr = bitsPtr;
    int leftSize = size;

    while (leftSize && rightSize)
    {
        if (std::toupper(*ptr) < std::toupper(*str))
        {
            return -1;
        }
        if (std::toupper(*ptr) > std::toupper(*str))
        {
            return 1;
        }
        leftSize--;
        rightSize--;
        ptr++;
        str++;
    }

    if (leftSize < rightSize)
    {
        return -1;
    }
    else if (leftSize > rightSize)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}


// split the string into two.  The current string becomes the front,
// which has all the characters w/ positions < p, while the new string
// (returned as a result) has the positions >= p.  Note that this is
// special in that it does NOT write to any of the string, so the
// underlying buffer is never written.  Handy, that.

INLINE_ void
CWBuffer::split(int p, CWBuffer* ptr)
{
    assert((p >= 0) && (p <= size));

    if (p == size)
    {
        // do not actually split, but return an empty CWBuffer.  p ==
        // size is a border case where we do nothing, and return an
        // empty string, since nothing belongs in the second half of
        // the split.

        ptr->clear();
    }
    else
    {
        // do the split

        *ptr = *this;

        // correct the new ptr
        ptr->bitsPtr = ptr->bitsPtr + p;
        ptr->size = ptr->size - p;
        ptr->allocSpace = ptr->allocSpace - p;

        // correct the old ptr
        allocSpace = p;
        size = p;
    }
}


// this function truncates a buffer by removing front characters from
// the front and back characters from the back of the buffer.
// it does not modify the underlying data

INLINE_ void
CWBuffer::truncate(int front, int back)
{
    assert((front + back) <= size);

    bitsPtr = bitsPtr + front;
    size = size - (front + back);
}



// basically just another way of specifying a truncation.  in this one,
// you specify the first and last characters of the string to keep.
// the rest are removed.

INLINE_ void
CWBuffer::substr(int first, int last)
{
    assert((first <= last) &&
           (first <= size) &&
           (last <= size) &&
           (first >= 0) &&
           (last >= 0));

    bitsPtr = bitsPtr + first;
    size = (last - first);
}


INLINE_ void
CWBuffer::makeWritable(int length)
{
    if (bufPtr)
    {
        if ( (length > allocSpace) || !bufPtr->mutex.compare(1))
        {
            // if we need more space than we have in this buffer, OR
            // if this is not the only CWBuffer pointing to this
            // CharArray, we need to copy it

            if (length > allocSpace)
            {
                allocSpace = length + 128;
            }

            CWCharArray* tmp = new CWCharArray(bitsPtr, size, allocSpace);
            bitsPtr = tmp->ptr;
            // get rid of the old pointer
            decrementCA(bufPtr);
            bufPtr = tmp;
            bufPtr->mutex.increment();
        }
    }
    else if (length > 0)
    {
        assert(size == 0);
        // allocate a buffer
        allocSpace = length + 128;

        bufPtr = new CWCharArray(allocSpace);
        bitsPtr = bufPtr->ptr;

        bufPtr->mutex.increment();
    }
}


INLINE_ void
CWBuffer::makeCString()
{
    if (bitsPtr != emptyStr)
    {
        makeWritable(size + 1);
        bitsPtr[size] = '\0';
    }
}


#endif
