#ifndef SPTR_HXX_
#define SPTR_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const SptrHeaderVersion =
    "$Id: Sptr.hxx,v 1.25 2000/11/02 19:59:01 veer Exp $";




#define SPTR_DEBUG 1
#define SPTR_TRACEBACK 1
#include <assert.h>
#include "VMutex.h"
#include "debug.h"

#if SPTR_DEBUG

#include <map>

/// this is a debugging class for pointers

class SptrDebugger
{
    private:
        SptrDebugger();

        static SptrDebugger* getInstance();

        typedef map < void*, bool > PtrMap;

        static SptrDebugger* impl_;

        PtrMap ptrsInUse;

        VMutex mutex;

    public:
        static void newPtr(void* ptr);

        static void deletePtr(void* ptr);
};


void sptrDebugMarkTraceback(void* sptr, void* ptr);
void sptrDebugClearTraceback(void* sptr, void* ptr);
void sptrDebugDumpTraceback(char* filename);

#endif


/** Template simulates a pointer with a "smart" reference that deletes
    the item when no more references exist
*/
template < class T >
class Sptr
{
public:

       mutable  T* ptr;
        mutable int* count;
        mutable VMutex* mutex;

        //    pthread_mutex_t* mutex;
private:
        void increment()
        {
            if (ptr)
            {
                if (!count)
                {
                    count = new int(0);
                }
                if (!mutex)
                {
                    mutex = new VMutex;
                }

                mutex->lock();
                (*count)++;
                mutex->unlock();
            }
#if SPTR_TRACEBACK
            sptrDebugMarkTraceback((void*) this, (void*) ptr);
#endif

        }

        void decrement()
        {
            if (ptr && count)
            {
#if SPTR_TRACEBACK
                sptrDebugClearTraceback((void*) this, (void*) ptr);
#endif
                bool countIsZero;
                mutex->lock();
                (*count)--;
                countIsZero = (*count == 0);
                mutex->unlock();
                if (countIsZero)
                {
#if SPTR_DEBUG
#if 0
                    cerr << "deleting " << ptr << endl;
                    breakpoint();
#endif
                    SptrDebugger::deletePtr((void*) ptr);
#endif
                    delete ptr; ptr = 0;
                    delete count; count = 0;
                    delete mutex; mutex = 0;
                }
            }
            ptr = 0;
            count = 0;
            mutex = 0;
        }

    public:
#if 0
        template < class T2 >
        operator Sptr<T2 > ()
        {
            return Sptr < T2 > (ptr, count, mutex);
        }
#endif
        Sptr() : ptr(0), count(0), mutex(0)
        {}
        ;

        Sptr(T* original, int* myCount = 0, VMutex* myMutex = 0)
        : ptr(original), count(myCount), mutex(myMutex)
        {
            if (ptr)
            {
#if SPTR_DEBUG
                if (!myCount || !myMutex)
                {
                    SptrDebugger::newPtr((void*) ptr);  // void* takes away warning
                }

#endif
                increment();
            }
        };

        template <class T2> Sptr(const Sptr<T2>& x)
        : ptr(x.ptr), count(x.count), mutex(x.mutex)
        {
            increment();
        }

        Sptr(const Sptr& x)
        : ptr(x.ptr), count(x.count), mutex(x.mutex)
        {
            increment();
        };

        ~Sptr()
        {
            {
                decrement();
            }
        }

        T& operator*() const
        {
            assert(ptr);
            return *ptr;
        }

        int operator!() const
        {
            if (ptr)
            {
                return (ptr == 0);
            }
            else
                return true;
        }

        T* operator->() const
        {
            return ptr;
        }

        template < class T2 > Sptr& dynamicCast(const Sptr < T2 > & x)
        {
            if (ptr == x.getPtr()) return *this;
            decrement();
            if(T* p = dynamic_cast < T* > (x.getPtr()))
            {
                count = x.getCount();
                mutex = x.getMutex();
                ptr = p;
                increment();
            }
            return *this;
        }

        template < class T2 >
        Sptr& operator=(const Sptr < T2 > & x)
        {
            if (ptr == x.getPtr()) return * this;
            decrement();
            ptr = x.getPtr();
            count = x.getCount();
            mutex = x.getMutex();
            increment();
            return *this;
        }


        Sptr& operator=(T* original)
        {
            if (ptr == original) return * this;
            decrement();
#if SPTR_DEBUG
            SptrDebugger::newPtr((void*) original);
#endif
            ptr = original;
            increment();
            return *this;
        };
        Sptr& operator=(const Sptr& x)
        {
            if (ptr == x.getPtr()) return * this;
            decrement();
            ptr = x.ptr;
            count = x.count;
            mutex = x.mutex;
            increment();
            return *this;
        }

	int operator != (const Sptr& rhs)
	{
		return (ptr != rhs.ptr);
	}

        friend int operator!=(const void* y, const Sptr& x)
        {
            if (x.ptr != y)
                return true;
            else
                return false;
        }

        friend int operator!=(const Sptr& x, const void* y)
        {
            if (x.ptr != y)
                return true;
            else
                return false;
        }

        friend int operator==(const void* y, const Sptr& x)
        {
            if (x.ptr == y)
                return true;
            else
                return false;
        }

        friend int operator==(const Sptr& x, const void* y)
        {
            if (x.ptr == y)
                return true;
            else
                return false;
        }

        int operator==(const Sptr& x) const
        {
            if (x.ptr == ptr)
                return true;
            else
                return false;
        }

        // the following interfaces are here because they may sometimes be
        // necessary.  DO NOT USE unless you are sure of what you are doing.

        VMutex* getMutex() const
        {
            return mutex;
        }

        int* getCount() const
        {
            return count;
        }

        T* getPtr() const
        {
            return ptr;
        }

};


#endif
