/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Poll_cxx_Version = 
    "$Id: Poll.cxx,v 1.15 2000/11/02 02:07:44 icahoon Exp $";


#include "Poll.hxx"
#include "Protocol.hxx"
#include "TransportCommon.hxx"
#include "SystemException.hxx"
#include "SystemStatus.hxx"
#include "ConnectionBrokenException.hxx"
#include "ProtocolException.hxx"
#include "IPAddress.hxx"
#include "Lock.hxx"
#include "VLog.hxx"
#include <cerrno>
#include <iomanip>



using namespace Vocal;
using namespace Vocal::Transport;
using Vocal::IO::Pipe;
using Vocal::Process::Lock;
using Vocal::Logging::VLog;


Poll::Poll()
    :	protocolsChanged_(false),
	interruptCount_(0)
{
    interruptor_.writeFD().setBlocking();
    interruptor_.readFD().setNonblocking();
    
    // Add interruptor to poll file descriptor list. We are marking the
    // protocols and unchanged, so that they won't get reloaded.
    //
    addInterruptorToFds();
}


Poll::~Poll()
{
}


void	    	    
Poll::registerProtocol(Protocol & protocol)
{
    const string    fn("Poll::registerProtocol");
    VLog    	    log(fn);

    pollfd  	newPollFd;
    
    newPollFd.fd = protocol.getFileDescriptor().getFD();
    newPollFd.events = newPollFd.revents = 0;
    
    string  	eventNames;
    bool    	awaitingSomething = false;
    
    if ( protocol.awaitIncoming() )
    {
    	newPollFd.events |= POLLIN;
	eventNames += " POLLIN";
	awaitingSomething = true;
    }
    if ( protocol.awaitOutgoing() )
    {
    	newPollFd.events |= POLLOUT;
	eventNames += " POLLOUT";
	awaitingSomething = true;
    }
    if ( protocol.awaitPriority() )
    {
    	newPollFd.events |= POLLPRI;
	eventNames += " POLLPRI";
	awaitingSomething = true;
    }
    
    if ( !awaitingSomething )
    {
    	VDEBUG(log) << fn << ": not awaiting any notifications, unregistering."
	    	    << VDEBUG_END(log);
		    
    	unregisterProtocol(protocol);
	return;
    }
        
    ProtocolMap::iterator   it = protocols_.find(newPollFd);
    
    if ( it != protocols_.end() )
    {
    	// This may be an update. If necessary, replace the pollfd, 
	// effectively reseting the events.
	//
	if ( (*it).first.events != newPollFd.events )
	{
	    // The whole pollfd is being used as a key in the pair, but 
	    // looking more closely, only the file descriptor is being 
	    // keyed off of. If we are here, the file descriptors 
	    // (i.e. the true keys) are the same, but the events may 
	    // be different. So casting the key portion of the pair to 
	    // a non const pollfd is ok.
	    //
	    // This is the penalty I am paying for storing data in
	    // the key.
	    //
	    pollfd & pollFd = const_cast<pollfd &>((*it).first);
            pollFd = newPollFd;
	    protocolsChanged_ = true;
	}
    }
    else 
    {
    	// The is a first time registration. Add the poll file
	// descriptor / protocol pair to the list.
	//
    	protocols_[newPollFd] = &protocol;
	protocolsChanged_ = true;
    }

    VDEBUG(log) << fn << "\n\t" << protocol 
    	    	<< "\n\tprotocols changed = " << ( protocolsChanged_ ? "true" : "false" )
    	    	<< "\n\tevents: " << eventNames
    	    	<< VDEBUG_END(log);

    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
}


void	    	    
Poll::updateProtocol(Protocol & protocol)
{
    registerProtocol(protocol);
}


void	    	    
Poll::unregisterProtocol(Protocol & protocol)
{
    const string    fn("Poll::unregisterProtocol");
    VLog    	    log(fn);

    pollfd  	pollFd;

    pollFd.fd = protocol.getFileDescriptor().getFD();

    bool    removed = false;
        
    ProtocolMap::iterator   it = protocols_.find(pollFd);
    if ( it != protocols_.end() )
    {
    	// Found it, so let's remove it.
	//
    	protocols_.erase(it);
	protocolsChanged_ = true;
	removed = true;
    }

    VDEBUG(log) << fn << "\n\t" << protocol 
    	    	<< "\n\tprotocols changed = " << ( protocolsChanged_ ? "true" : "false" )
		<< "\n\tremoved = " << ( removed ? "true" : "false" )
    	    	<< VDEBUG_END(log);

    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
}


int 	    	    
Poll::poll(int timeout)
throw ( Vocal::SystemException )
{
    const string    fn("Poll::poll");
    VLog    	    log(fn);
    
    // Recreate the poll file descriptor if needed.
    //
    if ( protocolsChanged_ )
    {
    	VDEBUG(log) << fn << ": protocols changed, updating." << VDEBUG_END(log);
	
    	fds_.clear();

	addInterruptorToFds();
	addProtocolsToFds();

	protocolsChanged_ = false;
    }

    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);

    VDEBUG(log) << fn << ": polling for activity. timeout = " << timeout
    	    	<< VDEBUG_END(log);
    
    int numberFdsActive = ::poll(&fds_[0], fds_.size(), timeout);

    if ( numberFdsActive < SUCCESS )
    {
    	int error = errno;

    	// If we were interrupted, treat it like a timeout.
	//
    	if ( error == EINTR )
	{
	    VDEBUG(log) << fn << ": interrupted." << VDEBUG_END(log);
    	    return ( 0 );
	}
	
    	throw Vocal::SystemException(fn + " on poll(): " + strerror(error), 
	    	    	    	 __FILE__, __LINE__, error);
    }    

    // Clear interruptor so that numberFdsActive corresponds to
    // registered file descriptors only.
    //
    for (   vector<pollfd>::iterator  it = fds_.begin();
    	    it != fds_.end() && numberFdsActive > 0;
	    it++
	)
    {
    	pollfd	    &   pollFdEntry = (*it);

    	if ( processInterruptor(pollFdEntry) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
	    pollFdEntry.revents = 0;
	    break;
	}
    }
        
    return ( numberFdsActive );
}


void 	    	    
Poll::processProtocols(int numberFdsActive)
throw ( Vocal::Transport::ProtocolException )
{
    const string    fn("Poll::processProtocols");
    VLog    	    log(fn);

    VDEBUG(log) << fn << ": Start: number file descriptors active = " 
    	    	<< numberFdsActive 
	    	<< ", protocols changed = " 
		<< ( protocolsChanged_ ? "true" : "false" )
		<< VDEBUG_END(log);

    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
		
    for (   vector<pollfd>::iterator  it = fds_.begin();
    	    it != fds_.end() && numberFdsActive > 0;
	    it++
	)
    {
    	pollfd	    &   pollFdEntry = (*it);

    	if ( pollFdEntry.fd == interruptor_.readFD().getFD() )
	{
	    pollFdEntry.revents = 0;
	    continue;
	}

    	ProtocolMap::iterator   protocolIt = protocols_.find(pollFdEntry);
	
	if ( protocolIt == protocols_.end() )
	{
	    // If a two protocols are active, but then second protocol
	    // is unregistered as a result of activity on the first 
	    // protocol, we will end up here.
	    //
	    protocolsChanged_ = true;

	    VDEBUG(log) << fn << ": no protocol found for: fd = "
	    	    	<< (*it).fd << VDEBUG_END(log);


    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    pollFdEntry.revents = 0;
	    continue;
	}

    	Protocol    * protocol = (*protocolIt).second;
    	assert( protocol != 0 );
	
    	if ( processPriority(pollFdEntry, protocol) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
	    pollFdEntry.revents = 0;
	    continue;
	}
	
    	if ( processIncoming(pollFdEntry, protocol) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
	    pollFdEntry.revents = 0;
	    continue;
	}
	
    	if ( processOutgoing(pollFdEntry, protocol) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
	    pollFdEntry.revents = 0;
	    continue;
	}
	
    	if ( processHangup(pollFdEntry, protocol) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
    	    pollFdEntry.revents = 0;
	    continue;
	}
	
    	if ( processError(pollFdEntry, protocol) )
	{
    	    VVERBOSE(log) << fn << ": " << *this << VVERBOSE_END(log);
	    numberFdsActive--;
	}

    	pollFdEntry.revents = 0;
    }

    VDEBUG(log) << fn << ": End: number file descriptors active = " 
    	    	<< numberFdsActive 
	    	<< ", protocols changed = " 
		<< ( protocolsChanged_ ? "true" : "false" )
		<< VDEBUG_END(log);
}


void	    	    
Poll::interrupt()
throw ( Vocal::SystemException )
{
    const string    fn("Poll::interrupt");
    VLog    	    log(fn);
    
    Lock    lock(interruptorMutex_); (void)lock;

    interruptCount_++;
    
    interruptor_.writeFD().write(&interruptCount_, 1);

    VDEBUG(log) << fn << ": Interrupting: "   << (unsigned)interruptCount_ 
    	    	<< " outstanding interrupts." << VDEBUG_END(log);
}


u_int8_t    	    
Poll::interruptCount() const
{
    return ( interruptCount_ );
}


ostream &           
Poll::writeTo(ostream & out) const
{
    out << "protocols (size: " << protocols_.size() << ") = \n{\n";
    for ( ProtocolMap::const_iterator it = protocols_.begin(); it != protocols_.end(); it++ )
    {
    	if ( (*it).first.fd == interruptor_.readFD().getFD() )
	{
	    continue;
	}
	
    	out << "  ";
	if ( (*it).second )
	{
	    out << *(*it).second << "\n";
	}
	else
	{
	    out << "Protocol Removed.\n";
	}
    }

    out << "}\ninterruptor = " << interruptor_ 
    	<< "\nnumber outstanding interruptions = " << (unsigned)interruptCount_
	<< "\nprotocolsChanged = " << ( protocolsChanged_ ? "true" : "false" );

    size_t  size = fds_.size();
    
    out << "\nfds (size: " << size << ") = \n{\n";
    for ( size_t i = 0; i < size; i++ )
    {
    	out << " ( " << fds_[i].fd 
	    << ", 0x" << hex << setw(2) << setfill('0') << (unsigned)(fds_[i].events)
	    << ", 0x" << hex << setw(2) << setfill('0') << (unsigned)(fds_[i].revents)
	    << dec << " )\n";
    }
    out << "}";
    
    return ( out );
}
					
void	    	    
Poll::addInterruptorToFds()
{
    pollfd  	interruptorPollFd;

    interruptorPollFd.fd    	= interruptor_.readFD().getFD();
    interruptorPollFd.events	= POLLIN;
    interruptorPollFd.revents	= 0;

    fds_.insert(fds_.end(), interruptorPollFd);
}


void	    	    
Poll::addProtocolsToFds()
{
    for (   ProtocolMap::iterator  it = protocols_.begin();
    	    it != protocols_.end();
	    it++
	)
    {
    	fds_.insert(fds_.end(), (*it).first);
    }
}


bool
Poll::processInterruptor(pollfd &   pollFdEntry)
{
    const string    fn("Poll::processInterruptor");
    VLog    	    log(fn);

    VVERBOSE(log) << fn << ": fd = " << pollFdEntry.fd << VVERBOSE_END(log);
        
    if  ( pollFdEntry.fd == interruptor_.readFD().getFD() )
    {
    	if ( pollFdEntry.revents & POLLIN  )
	{
	    u_int8_t    interruptNumber = 0;
	    IPAddress   remoteAddr;

	    Lock    lock(interruptorMutex_); (void)lock;

	    VDEBUG(log) << fn << ": Interrupted: " << (unsigned)interruptCount_ 
    	    		<< " outstanding interrupts." << VDEBUG_END(log);
	    do
	    {
		// Let the user catch the exception, if thrown.
		//
        	interruptor_.readFD().read(&interruptNumber, 1);

	    } while ( interruptNumber < interruptCount_ );

	    interruptCount_ = 0;
	    
	    return ( true );
	}
	else
	{
	    return ( false );
	}
    }
    return ( false );
}


bool
Poll::processPriority(pollfd & pollFdEntry, Protocol * protocol)
{
    const string    fn("Poll::processPriority");
    VLog    	    log(fn);

    VDEBUG(log) << fn << ": fd = " << pollFdEntry.fd << VDEBUG_END(log);

    if ( pollFdEntry.revents & POLLPRI )
    {
	VDEBUG(log) << fn << ": Priority for protocol = " << *protocol
	    	    << VDEBUG_END(log);

    	// Notify the user of a priority message on the file descriptor.
	//
	try
	{
	    protocol->onPriority();
	}
	catch ( Vocal::SystemStatus & )
	{
	    // The status indicates that we need to poll again.
	    // This is the desfault behavior, so do nothing.
	}
	catch ( Vocal::Transport::ConnectionBrokenException & e)
	{
	    return ( processHangup(pollFdEntry, protocol, &e) );
	}
	catch ( Vocal::SystemException & e )
	{
	    return ( processError(pollFdEntry, protocol, &e) );
	}
	catch ( ... )
	{
	    protocol->handleUnknownException();
	}
	return ( true );
    }
    
    return ( false );
}


bool
Poll::processIncoming(pollfd & pollFdEntry, Protocol * protocol)
{
    const string    fn("Poll::processIncoming");
    VLog    	    log(fn);

    VVERBOSE(log) << fn << ": fd = " << pollFdEntry.fd << VVERBOSE_END(log);

    if ( pollFdEntry.revents & POLLIN )
    {
	VDEBUG(log) << fn << ": IncomingAvailable for protocol = " 
		    << *protocol << VDEBUG_END(log);

    	// Notify the user of an incoming message on the file 
	// descriptor.
	//
	try
	{
	    protocol->onIncomingAvailable();
	}
	catch ( Vocal::SystemStatus & )
	{
	    // The status indicates that we need to poll again.
	    // Do nothing.
	}
	catch ( Vocal::Transport::ConnectionBrokenException & e )
	{
	    return ( processHangup(pollFdEntry, protocol, &e) );
	}
	catch ( Vocal::SystemException & e )
	{
	    return ( processError(pollFdEntry, protocol, &e) );
	}
	catch ( ... )
	{
	    protocol->handleUnknownException();
	}
	return ( true );
    }
    return ( false );
}


bool
Poll::processOutgoing(pollfd & pollFdEntry, Protocol * protocol)
{
    const string    fn("Poll::processOutgoing");
    VLog    	    log(fn);

    VVERBOSE(log) << fn << ": fd = " << pollFdEntry.fd << VVERBOSE_END(log);

    if ( pollFdEntry.revents & POLLOUT )
    {
	VDEBUG(log) << fn << ": OutgoingAvailable for protocol = " 
		    << *protocol << VDEBUG_END(log);

    	// Notify the user of that the file descriptor is
	// available to send outgoing messages.
	//
	try
	{
	    protocol->onOutgoingAvailable();
	}
	catch ( Vocal::SystemStatus )
	{
	    // The status indicates that we need to poll again.
	    // Do nothing.
	}
	catch ( Vocal::Transport::ConnectionBrokenException & e )
	{
	    return ( processHangup(pollFdEntry, protocol, &e) );
	}
	catch ( Vocal::SystemException & e )
	{
	    return ( processError(pollFdEntry, protocol, &e) );
	}
	catch ( ... )
	{
	    protocol->handleUnknownException();
	}
	return ( true );
    }
    return ( false );
}


bool
Poll::processHangup(
    pollfd  	    	    	&   pollFdEntry, 
    Protocol 	    	    	*   protocol,
    ConnectionBrokenException 	*   connectionBrokenException
)
{
    const string    fn("Poll::processHangup");
    VLog    	    log(fn);

    VVERBOSE(log) << fn << ": fd = " << pollFdEntry.fd << VVERBOSE_END(log);
    
    if  (   connectionBrokenException != 0
    	||  pollFdEntry.revents & POLLHUP 
	)
    {
	VDEBUG(log) << fn << ": Disconnect for protocol = " 
		    << *protocol << VDEBUG_END(log);

    	// Notify the user of that the file descriptor has hung up.
	//
	try
	{
	    protocol->onDisconnect(connectionBrokenException);
	}
	catch ( Vocal::SystemStatus & )
	{
	    // The status indicates that we need to poll again.
	    // Do nothing.
	}
	catch ( Vocal::SystemException & e )
	{
	    return ( processError(pollFdEntry, protocol, &e) );
	}
	catch ( ... )
	{
	    protocol->handleUnknownException();
	}
	return ( true );
    }
    return ( false );
}


bool
Poll::processError(
    pollfd  	    	&   pollFdEntry, 
    Protocol 	    	*   protocol,
    SystemException	*   systemException
)
{
    const string    fn("Poll::processError");
    VLog    	    log(fn);

    VVERBOSE(log) << fn << ": fd = " << pollFdEntry.fd << VVERBOSE_END(log);

    if  (   systemException != 0 
    	||  pollFdEntry.revents & POLLERR
	||  pollFdEntry.revents & POLLNVAL
	)
    {
	VWARN(log)  << fn << ": Error for protocol = " << *protocol
	    	    << VWARN_END(log);

    	// Notify the user of the error.
	//
	try
	{
            protocol->onError(systemException);
	}
	catch ( ... )
	{
	    protocol->handleUnknownException();
	}
	return ( true );
    }
    return ( false );
}



bool 
pollfdLess::operator()(const pollfd & left, const pollfd & right) const
{
    return ( left.fd < right.fd );
}
