
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Thread_cxx_Version = 
    "$Id: Thread.cxx,v 1.11 2000/10/26 02:41:42 icahoon Exp $";


#include <assert.h>
#include "Thread.hxx"
#include "Runnable.hxx"
#include "SignalHandler.hxx"
#include "Application.hxx"
#include "Lock.hxx"
#include <stdio.h>


using Vocal::Process::Thread;
using Vocal::Process::Runnable;
using Vocal::Process::SignalHandler;
using Vocal::Process::Lock;
using Vocal::ReturnCode;


Thread::ThreadMap   		*   Thread::threadMap_ = 0;
VMutex		    	    	    Thread::threadMapMutex_;
Thread          	    	*   Thread::mainThread_ = 0;


void	    	    
Thread::init()
{
    Lock    lock(threadMapMutex_); (void)lock;
    
    if ( !threadMap_ )
    {
        threadMap_ = new map<vthread_t, Thread *>;
        mainThread_ = new Thread;

    	vthread_t   mainThreadId = VThread::selfId();
	
	(*threadMap_)[mainThreadId] = mainThread_;
    }
}


void	    	    
Thread::uninit()
{
    Lock    lock(threadMapMutex_); (void)lock;

    delete threadMap_;
    threadMap_ = 0;

    delete mainThread_;
    mainThread_ = 0;
}


Thread *
Thread::self()
{
    init();
    
    Lock    lock(threadMapMutex_); (void)lock;

    vthread_t	id = VThread::selfId();

    assert( threadMap_ != 0 );
    
    ThreadMap::iterator it = threadMap_->find(id);
    
    return ( it == threadMap_->end() ? 0 : (*it).second );
}


vthread_t
Thread::selfId()
{
    return ( VThread::selfId() );
}


Thread::Thread(Runnable & runnable, const char * name)
    :	thread_(new VThread),
    	runnable_(runnable),
	signalHandler_(0),
	name_(name ? name : "thread")
{
    init();

    thread_->spawn(threadMain, this);
}


Thread::~Thread()
{
    const vthread_t	*   threadId = thread_->getId();

    if ( threadId )
    {
        Lock    lock(threadMapMutex_); (void)lock;

    	if ( threadMap_ )
	{
            threadMap_->erase(*threadId);
	}
	else
	{
    	    assert( threadMap_ != 0 );
	}
    }

    delete thread_;
    
    delete signalHandler_;
}


ReturnCode  	    	    
Thread::join()
{
    // Can't join from your own thread.
    //
    if ( this == self() )
    {
    	return ( 0 );
    }
    
    ReturnCode rc = 0;
    
    if ( thread_ )
    {
    	rc = thread_->join();
    }
    return ( rc );
}


void
Thread::exit()
{
    // Can't exit unless this is the thread of execution.
    //
    if ( thread_ == 0 || this != self() )
    {
    	return;
    }

    thread_->exit();
}


Runnable &
Thread::runnable()
{
    return ( runnable_ );
}	


SignalHandler &     	    
Thread::signalHandler()
{
    assert( signalHandler_ != 0 );
    return ( *signalHandler_ );
}


ostream &           
Thread::writeTo(ostream & out) const
{
    return ( out << name_.c_str() );
}	


void *	    
Thread::threadMain(void * data)
{
    Thread *	thread = (Thread *)data;
    
    if ( thread == 0 )
    {
    	return ( (void *)1 );
    }
    
    const vthread_t	*   threadId = thread->thread_->getId();

    // Scope the lock for update of thread map only.
    //
    {
        Lock    lock(threadMapMutex_); (void)lock;
	if ( threadMap_ )
	{
    	    (*threadMap_)[*threadId] = thread;
	}
	else
	{
	    assert( threadMap_ != 0 );
	    return ( (void *)1 );
	}
    }

    // Update name
    //
    char	    buffer[64];
    sprintf(buffer, "%lu", (unsigned long)*threadId);
    thread->name_ += '.';
    thread->name_ += buffer;

    // The signalHandler will persist for the lifetime of the run.
    //
    thread->signalHandler_ = new SignalHandler;

    void    * rc = (void *)(thread->runnable_.run());

    delete thread->signalHandler_;
    thread->signalHandler_ = 0;
    
    return ( rc );
}


Thread::Thread()
    :	thread_(0),
    	runnable_(Application::instance()),
	signalHandler_(new SignalHandler),
	name_("main")
{
    char	    buffer[64];
    sprintf(buffer, "%lu", (unsigned long)VThread::selfId());
    name_ += '.';
    name_ += buffer;
}
